/* OperServ routines

   OperStats - Copyright (c) 2000-2001 Darcy Grexton
   Contact: cygnus@uniserve.ca, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#include "../inc/operstats.h"
#include "../inc/os-help.h"
#include "../inc/shared-help.h"

AutoKill *akilllist[1024];
MSG *msglist = NULL;
Trigger *triggerlist = NULL;
Exception *exceptionlist = NULL;

static void do_help (const char *source);
static void do_global (const char *source);
static void do_jupe (const char *source);
static void do_clones (const char *source);
static void do_akill (const char *source);
static void do_akill_add (const char *source);
static void do_akill_del (const char *source);
static void do_akill_list (const char *source);
static void do_logonmsg (const char *source);
static void do_logonmsg_add (const char *source);
static void do_logonmsg_del (const char *source);
static void do_logonmsg_list (const char *source);
static void do_trigger (const char *source);
static void do_trigger_add (const char *source);
static void do_trigger_del (const char *source);
static void do_trigger_list (const char *source);
static void do_exception (const char *source);
static void do_exception_add (const char *source);
static void do_exception_del (const char *source);
static void do_exception_list (const char *source);
static void sh_auth (const char *source);
static void sh_deauth (const char *source);
#ifdef RAWINJECT
static void sh_raw (const char *source);
static void sh_inject (const char *source);
#endif
static void sh_debug (const char *source);
static void sh_shutdown (const char *source);
static void sh_restart (const char *source);
static void sh_settings (const char *source);
static void sh_update (const char *source);
static void sh_rehash (const char *source);

/* Temporary debugging functions. */
static void do_userlist (const char *source);
static void do_chanlist (const char *source);

/* Add a new logonmsg: This is also called when loading the db */
static MSG *new_message (const char *nick, int set, const char *text)
{
    MSG *msg;

    msg = scalloc (sizeof (MSG), 1);

    msg->text = sstrdup (text);
    msg->setter = sstrdup (nick);
    msg->set = set;

    msg->next = msglist;

    if (msg->next)
	msg->next->prev = msg;

    msglist = msg;

    return msg;
}

/* Add a new trigger to the trigger list */
Trigger *new_trigger (const char *host, int limit)
{
    Trigger *trigger;
    char *username, *hostname;

    username = sstrdup (host);
    hostname = strchr (username, '@');

    if (!hostname)
	return NULL;

    *hostname++ = 0;

    if ((trigger = findtrig (username, hostname)))
    {
	free (username);
	return trigger;
    }

    trigger = scalloc (sizeof (Trigger), 1);

    strscpy (trigger->host, host, HOSTLEN);
    trigger->limit = limit;

    trigger->next = triggerlist;

    if (trigger->next)
        trigger->next->prev = trigger;

    triggerlist = trigger;

    free (username);

    return trigger;
}

/* Search for a trigger */
Trigger *findtrig (const char *user, const char *host)
{
    Trigger *trigger;
    char *mask;

    mask = smalloc (strlen (user) + strlen (host) + 2);
    sprintf (mask, "%s@%s", user, host);

    for (trigger = triggerlist; trigger; trigger = trigger->next)
	if (match_wild_nocase (trigger->host, mask))
	{
	    free (mask);
	    return trigger;
	}

    free (mask);
    return NULL;
}

/* Add a new exception to the exception list */
Exception *new_exception (const char *host)
{
    Exception *exception;
    char *username, *hostname;

    username = sstrdup (host);
    hostname = strchr (username, '@');

    if (!hostname)
	return NULL;

    *hostname++ = 0;

    if ((exception = findexcept (username, hostname)))
    {
	free (username);
	return exception;
    }

    exception = scalloc (sizeof (Exception), 1);

    strscpy (exception->host, host, HOSTLEN);

    exception->next = exceptionlist;

    if (exception->next)
	exception->next->prev = exception;

    exceptionlist = exception;

    free (username);

    return exception;
}

/* Search for an exception */
Exception *findexcept (const char *user, const char *host)
{
    Exception *exception;
    char *mask;

    mask = smalloc (strlen (user) + strlen (host) + 2);
    sprintf (mask, "%s@%s", user, host);

    for (exception = exceptionlist; exception; exception = exception->next)
	if (match_wild_nocase (exception->host, mask))
	{
	    free (mask);
	    return exception;
	}

    free (mask);

    return NULL;
}   

/* Main OperServ routine. */
void operserv (const char *source, char *buf)  
{
    char *cmd, *s;
    char orig[BUFSIZE];

    /* Make a copy of the original message. */ 
    strscpy (orig, buf, sizeof (orig));   

    cmd = strtok (buf, " ");

    if (!cmd)
	return;
    
    if (!stricmp (cmd, "\1PING"))
    {
	if (!(s = strtok (NULL, "")))
	    s = "0";
	strip (s);
	ctcpreply (s_OperServ, source, "PING %s", s);
    }
    else if (!stricmp (cmd, "\1VERSION\1"))
	ctcpreply (s_OperServ, source, "VERSION OperStats v%s", version);

    /* A CTCP we don't recognize.. Ignore it. */
    else if (*cmd == '\1')
	return;

    else if (!is_oper (source))
    {
	notice (s_OperServ, source, OS_ACCESS_DENIED);
	return;
    }

    else
    {
	Hash *command, hash_table[] =
	{
	    {"HELP",		H_IRCOP,	do_help},
	    {"GLOBAL",		H_IRCOP,	do_global},
	    {"JUPE",		H_IRCOP,	do_jupe},
	    {"CLONES",		H_IRCOP,	do_clones},
	    {"AKILL",		H_IRCOP,	do_akill},
	    {"TRIGGER",		H_IRCOP,	do_trigger},
	    {"EXCEPTION",	H_IRCOP,	do_exception},
	    {"AUTH",		H_IRCOP,	sh_auth},
	    {"DEAUTH",		H_SRA,		sh_deauth},
	    {"SHUTDOWN",	H_SRA,		sh_shutdown},
	    {"RESTART",		H_SRA,		sh_restart},
#ifdef RAWINJECT
	    {"RAW",		H_SRA,		sh_raw},
	    {"INJECT",		H_SRA,		sh_inject},
#endif
	    {"DEBUG",		H_SRA,		sh_debug},
	    {"UPDATE",		H_SRA,		sh_update},
	    {"REHASH",		H_SRA,		sh_rehash},
	    {"SETTINGS",	H_SRA,		sh_settings},
	    {"LOGONMSG",	H_SRA,		do_logonmsg},
	    {"USERLIST",	H_SRA,		do_userlist},
	    {"CHANLIST",	H_SRA,		do_chanlist},
	    {NULL}
	};
            
	if ((command = get_hash (s_OperServ, source, strupper (cmd),
				 hash_table)))
	{
	    /* Check if they want a disabled command. */
	    if ((command->process == do_global &&
		globalnoticer_on == FALSE) ||
		(command->process == do_logonmsg &&
		globalnoticer_on == FALSE) ||
		(command->process == do_clones &&
		!maxclones) ||
		(command->process == do_trigger &&
		!maxclones) ||
		(command->process == do_exception &&
		!maxclones))
	    {
		notice (s_OperServ, source, ERR_UNKNOWN_COMMAND,
		    strupper (cmd), s_OperServ);
		return;
	    }

	    (*command->process) (source);

	    if (!(command->process == sh_auth)) /* No pw snooping */
	    {
		log ("OperServ: <%s> %s", source, orig);
		snoop (s_OperServ, "<%s> %s", source, orig);
	    }
	    else /* But we do want to snoop something. */
	    {
		log ("OperServ: <%s> AUTH ...", source);
		snoop (s_OperServ, "<%s> AUTH ...", source);
	    }
	}
    }
}

/* Return a help message. */
static void do_help (const char *source) 
{
    char *cmd = strtok (NULL, " ");
            
    if (!cmd)
    {
	notice_list (s_OperServ, source, os_help_start);
	notice_list (s_OperServ, source, os_base_help);
	if (globalnoticer_on == TRUE)
	    notice_list (s_OperServ, source, os_help_global);
	notice_list (s_OperServ, source, shared_oper_help);
	if (is_sra (source))
	{
	    notice_list (s_OperServ, source, shared_sra_help);
	    if (globalnoticer_on == TRUE)
		notice_list (s_OperServ, source, os_help_logonmsg);
	}
	notice_list (s_OperServ, source, os_help_end);
    }
    else
    {
	Hash_HELP *command, hash_table[] =
	{
	    {"GLOBAL",		H_IRCOP,	global_help},
	    {"JUPE",		H_IRCOP,	jupe_help},
	    {"CLONES",		H_IRCOP,	clones_help},
	    {"AKILL",		H_IRCOP,	akill_help},
	    {"TRIGGER",		H_IRCOP,	trigger_help},
	    {"EXCEPTION",	H_IRCOP,	exception_help},
	    {"AUTH",		H_IRCOP,	auth_help},
	    {"DEAUTH",		H_SRA,		deauth_help},
	    {"SHUTDOWN",	H_SRA,		shutdown_help},
	    {"RESTART",		H_SRA,		restart_help},
#ifdef RAWINJECT
	    {"RAW",		H_SRA,		raw_help},
	    {"INJECT",		H_SRA,		inject_help},
#endif
	    {"DEBUG",		H_SRA,		debug_help},
	    {"UPDATE",		H_SRA,		update_help},
	    {"LOGONMSG",	H_SRA,		logonmsg_help},
	    {"REHASH",		H_SRA,		rehash_help},
	    {"SETTINGS",	H_SRA,		settings_help},
	    {NULL}
	};

	if ((command = get_help_hash (s_OperServ, source, strupper (cmd),
				      hash_table)))
	{
	    /* Check if they want a disabled command. */
	    if ((command->process == global_help &&
		globalnoticer_on == FALSE) ||
		(command->process == logonmsg_help &&
		globalnoticer_on == FALSE) ||
		(command->process == clones_help &&
		!maxclones) ||
		(command->process == trigger_help &&
		!maxclones) ||
		(command->process == exception_help &&
		!maxclones))
	    {
		notice (s_OperServ, source, ERR_NO_HELP, strupper (cmd));
		return;
	    }
	    notice_list (s_OperServ, source, os_help_start);
	    notice_list (s_OperServ, source, command->process);
	    notice_list (s_OperServ, source, os_help_end);
	}
    }
}

/* Jupiter a server or nickname. */
static void do_jupe (const char *source)
{
    char *param = strtok (NULL, " ");
    char *reason = strtok (NULL, "");

    if (!param)
    {
	notice (s_OperServ, source, RPL_SYNTAX, "JUPE Server|Nick [Reason]");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "JUPE");
	return;
    }

    /* Sanity checks .. We could do these all in one shot, but the output
       (and resultant warnings) are more informative this way.
     */
    if (!stricmp (me.name, param))
    {
	notice (s_OperServ, source, ERR_NOT_OPERSTATS);
	globops (s_OperServ, OS_JUPE_MY_SERVER, source);
	return;
    }

    if (!stricmp (me.ruplink, param))
    {
	notice (s_OperServ, source, ERR_NOT_OPERSTATS);
	globops (s_OperServ, OS_JUPE_MY_UPLINK, source);
	return;
    }

    if (is_one_of_mine (param))
    {
	notice (s_OperServ, source, ERR_NOT_OPERSTATS);
	globops (s_OperServ, OS_JUPE_MY_NICK, source, param);
	return;
    }

    /* Is it a server? */
    if (strchr (param, '.'))
    {
	int i;

	/* Does that server already exist? */
	for (i = 0; i < servcnt; i++)
	{
	    if (!stricmp (servlist[i].server, param))
	    {
		/* Yep. Now, we squit it from the network, and free the
		   memory for it. */
		send_cmd (me.name, "%s %s :Juped by %s: %s", me.token ? "-" : 
		    "SQUIT", param, source, reason ? reason :
		    "No reason given");
		--servcnt;
		--servstat;
		if (servlist[i].server)
		    free (servlist[i].server);
		if (servlist[i].uplink)
		    free (servlist[i].uplink);
		if (i < servcnt)
		    bcopy (servlist + i + 1, servlist + i,
			sizeof (*servlist) * (servcnt - i));
	    }
	}

	/* Now the server is either gone, or was never there, so we bring
	   on the jupe.
	 */
#ifdef IRCUP9
	send_cmd (NULL, "SERVER %s 2 %lu %lu P09 :%s - %s", param,
	    time (NULL), time (NULL), reason ? reason : "Jupitered Server",
	    source);
#else
	send_cmd (NULL, "%s %s 2 :%s - %s", me.token ? "'" : "SERVER",
	    param, reason ? reason : "Jupitered Server", source);
#endif /* IRCUP9 */
	globops (s_OperServ, OS_JUPED, param, source);
	notice (s_OperServ, source, OS_JUPE_SUCCESSFUL, param);
    }

    /* Doesn't appear to be a server, treat it as a nickname. */
    else
    {
	User *u;
	char buf[256];

	/* Is the nick in use? */
	if ((u = finduser (param)))
	{
	    /* Yep. Now we kill the client before bringing on the jupe. */
	    sprintf (buf, "Jupitered by %s: %s", source,
		reason ? reason : "No reason given");
	    kill_user (s_OperServ, u->nick, buf);
	}

	/* Now the nick is gone, or was never there, so we bring on the
	   jupe.
	 */
	send_cmd (NULL, "%s %s %ld 1 jupe %s %s :%s - %s", me.token ? "&" :
	    "NICK", param, time (NULL), os.host, me.name, reason ? reason :
	    "Jupitered Nick", source);
	globops (s_OperServ, OS_JUPED, param, source);
	notice (s_OperServ, source, OS_JUPE_SUCCESSFUL, param);
    }
}

/* Send a global notice */
static void do_global (const char *source)
{
    char *text = strtok (NULL, "");

    if (!text)
    {
	notice (s_OperServ, source, RPL_SYNTAX, "GLOBAL Message");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "GLOBAL");
	return;
    }

    noticeall (text);
    globops (s_OperServ, OS_GLOBAL_NOTICE, source);
}

/* Handle an AKILL command */
static void do_akill (const char *source)
{
    char *cmd = strtok (NULL, " ");

    if (!cmd)
    {
	notice (s_OperServ, source, RPL_SYNTAX,
	    "AKILL ADD|DEL|LIST [Mask|Number|Option]");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "AKILL");
    }
    else
    {
	Hash_SUB *command, hash_table[] =
	{
	    {"ADD",		H_IRCOP,	do_akill_add},
	    {"DEL",		H_IRCOP,	do_akill_del},
	    {"LIST",		H_IRCOP,	do_akill_list},
	    {NULL}
	};

	if ((command = get_sub_hash (s_OperServ, source, strupper (cmd),
				 hash_table)))
	    (*command->process) (source);
	else
	{
	    notice (s_OperServ, source, RPL_SYNTAX,
		"AKILL ADD|DEL|LIST [Mask|Number|Option]");
	    notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "AKILL");
	}
    }
}

/* Add an AKill to the AKill list */
void do_akill_add (const char *source)
{
    AutoKill *akill, **list;
    User *u;
    char *param, mask[BUFSIZE], reason[BUFSIZE], *tmp, orig[BUFSIZE], *username;
    char *hostname;
    int i, j = 0, nonwild = 0, realname = 0, perm = 0, expires;

    param = strtok (NULL, " ");
    tmp = strtok (NULL, "");

    if (!param)
    {
	notice (s_OperServ, source, RPL_SYNTAX,
	    "AKILL ADD Mask [Expiry] Reason");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "AKILL");
	return;
    }

    if (!tmp)
	memcpy (reason, def_akill_reason, sizeof (reason));
    else
	memcpy (reason, tmp, sizeof (reason));

    strscpy (orig, reason, sizeof (orig));

    /* If we don't find an expire time below, we'll use this default */
    expires = def_akill_time;

    tmp = strtok (reason, " ");

    if (!stricmp (tmp, "!P"))
    {
	expires = 0;
	j = 1;
	perm = 1;

	if (strstr (orig, "!R"))
	    tmp = strtok (NULL, " ");
    }

    if (!stricmp (tmp, "!R"))
    {
	realname = 1;
	j = 1;

	if (strstr (orig, "!T"))
	    tmp = strtok (NULL, " ");
    }

    if (!stricmp (tmp, "!T"))
    {
	tmp = strtok (NULL, " ");
	expires = dotime (tmp);
	j = 1;
	if (expires <= 0)
	    expires = def_akill_time;
    }

    if (!j)
	memcpy (reason, orig, sizeof (reason));
    else
	memcpy (reason, strtok (NULL, ""), sizeof (reason));

    /* Is the param a nick? If so, get the mask for it. */
    if ((u = finduser (param)))
	snprintf (mask, sizeof (mask), "*@%s", u->host);
    else
	strscpy (mask, param, BUFSIZE);

    if (mask && reason)
    {
	if (is_akilled (mask))
	{
	    notice (s_OperServ, source, ERR_LIST_THERE, mask, "", "AKill");
	    return;
	}

	/* Sanity checks - These don't apply if it's a realname akill */
	if ((strchr (mask, '!') && !realname) ||
	    (!strchr (mask, '@') && !realname) ||
	    (mask[strlen(mask)-1] == '@' && !realname))
	{
	    notice (s_OperServ, source, OS_AKILL_FORMAT, "");
	    return;
	}

	if (mask[0] == '@')
	{
	    char *mymask;
	    mymask = smalloc (strlen (mask) + 2);
	    strcpy (mymask, "*");
	    strcat (mymask, mask);
	    strcpy (mask, mymask);
	    free (mymask);
        }

	/* Make sure theres a sufficient number of non-wildcard characters.
	   @ and . are counted as 'wildcards' .. SRAs and realname akills
	   are exempt from this.
	 */
	if (!is_sra (source) && !realname)
	{
	    for (i = 0; i < strlen (mask); i++)
		if (!(mask[i] == '*' || mask[i] == '?' || mask[i] == '@' ||
		      mask[i] == '.'))
		    nonwild++;

	    if (nonwild < nonwildreq)
	    {
		notice (s_OperServ, source, ERR_TOOMANYWILD, nonwildreq);
		return;
	    }
	}

	/* Passed all the checks, now add it to the AKill list */
	akill = scalloc (sizeof (AutoKill), 1);
	akill->mask = sstrdup (mask);
	akill->reason = sstrdup (reason);
	akill->setter = sstrdup (source);
	akill->realname = realname;
	akill->set = time (NULL);
	akill->expires = expires;
	list = &akilllist[AHASH(akill->mask)];
	akill->next = *list;
	if (*list)
	    (*list)->prev = akill;
        *list = akill;

	notice (s_OperServ, source, OS_AKILL_ADDED, mask);
	if (globalonakill_on == TRUE)
	    globops (s_OperServ, OS_AKILL_ADDED_GLOBOP, source,
		realname ? "Realname " : "", perm ? "P" : "", mask, reason);

	if (!realname)
	{
	    username = sstrdup (akill->mask);
	    hostname = strchr (username, '@');

	    if (!hostname)
		return;

	    *hostname++ = 0;

#ifdef IRCUP9
	    /* We use 999999999 because we'll expire it ourselves. */
	    send_cmd (me.name, "GLINE * +*@%s 999999999 :%s",
		hostname, akill->reason);
#elif defined (BAHAMUT)
	    send_cmd (me.name, "AKILL %s * 0 %s %ld :%s", hostname,
		s_OperServ, time (NULL), akill->reason);
#else
	    send_cmd (me.name, "%s %s * :%s", me.token ? "V" : "AKILL",
		hostname, akill->reason);
#endif
	    free (username);
	}
	else
	{
	    int j = 0;

	    /* Hunt down any users that match this AKill. */
	    while (j < 1023)
	    {
		j++;
		u = userlist[j];
		if (u && match_wild_nocase (mask, u->real))
		{
#ifdef IRCUP9
		    /* We use 999999999 because we'll expire it ourselves. */
		    send_cmd (me.name, "GLINE * +*@%s 999999999 :%s",
			u->host, akill->reason);
#elif defined (BAHAMUT)
		    send_cmd (me.name, "AKILL %s * 0 %s %ld :%s", u->host,
			s_OperServ, time (NULL), akill->reason);
#else
		    send_cmd (me.name, "%s %s * :%s", me.token ? "V" :
			"AKILL", u->host, akill->reason);
#endif
		    u = u->next;
		}
	    }
	}
    }
    else
    {
	notice (s_OperServ, source, RPL_SYNTAX,
	    "AKILL ADD Mask [Expiry] Reason");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "AKILL");
    }
}

/* Delete an AKill */
static void do_akill_del (const char *source)
{
    AutoKill *akill;
    int i, j = 0;
    char *param = strtok (NULL, " ");
#ifndef IRCUP9
    char *user, *host;
#endif

    if (!param)
    {
	notice (s_OperServ, source, RPL_SYNTAX, "AKILL DEL Number|ALL");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "AKILL");
	return;
    }

    if (!stricmp (param, "ALL"))
    {
	for (i = 0; i < 1024; i++)
	{
	    for (akill = akilllist[i]; akill; akill = akill->next)
	    {
		j++;

		free (akill->setter);
		free (akill->reason);

#ifdef IRCUP9
		send_cmd (me.name, "GLINE * -%s", akill->mask);
#else
		user = sstrdup (akill->mask);
		host = strchr (user, '@');

		if (host)
		{
		    *host++ = 0;

		    send_cmd (me.name, "%s %s %s", me.token ? "Y" : "RAKILL",
			host, user);
		}

		free (user);
#endif /* IRCUP9 */

		if (akill->prev)
		    akill->prev->next = akill->next;
		else
		    akilllist[AHASH(akill->mask)] = akill->next;
		if (akill->next)
		    akill->next->prev = akill->prev;
	    }
	}

	free (akill);
	notice (s_OperServ, source, OS_ALL_AKILL_DELETED, j);
	return;
    }
    else
    {
	for (i = 0; i < 1024; i++)
	{
	    for (akill = akilllist[i]; akill; akill = akill->next)
	    {
		j++;

		if (j == atoi (param) || !stricmp (param, akill->mask))
		{
		    notice (s_OperServ, source, OS_AKILL_DELETED, akill->mask);

		    free (akill->setter);
		    free (akill->reason);

#ifdef IRCUP9
		    send_cmd (me.name, "GLINE * -%s", akill->mask);
#else
		    user = sstrdup (akill->mask);
		    host = strchr (user, '@');

		    if (host) 
		    {
			*host++ = 0;

			send_cmd (me.name, "%s %s %s", me.token ? "Y" :
			    "RAKILL", host, user);
		    }

		    free (user);
#endif /* IRCUP9 */

		    if (akill->prev)
			akill->prev->next = akill->next;
		    else
			akilllist[AHASH(akill->mask)] = akill->next;
		    if (akill->next)
			akill->next->prev = akill->prev;

		    free (akill);

		    return;
		}
	    }
	}
    }
    notice (s_OperServ, source, OS_AKILL_NOT_FOUND, param);
}

/* List all AKills, or a specific akill */
static void do_akill_list (const char *source)
{
    AutoKill *akill;
    int i, j = 0, showreasons = 0;
    char *param1 = strtok (NULL, " ");
    char *param2 = strtok (NULL, " ");

    /* We only show reasons if asked to. */
    if (param1 && !stricmp (param1, "FULL"))
	showreasons = 1;

    notice (s_OperServ, source, OS_AKILL_LIST,
	showreasons ? " - With reasons" : "",
	(showreasons && param2) ? " - Matching " :
	(!showreasons && param1) ? " - Matching " : "",
	(showreasons && param2) ? param2 : (!showreasons && param1) ?
	    param1 : "");

    for (i = 0; i < 1024; i++)
    {
	for (akill = akilllist[i]; akill; akill = akill->next)
	{
	    if (showreasons && param2)
		if (!match_wild_nocase (param2, akill->mask) &&
		    !match_wild_nocase (param2, akill->setter))
		    continue;

	    if (!showreasons && param1)
		if (!match_wild_nocase (param1, akill->mask) &&
		    !match_wild_nocase (param1, akill->setter))
		    continue;

	    j++;

	    if (akill->expires == 0)
		notice (s_OperServ, source, OS_PAKILL_LIST_MATCH, j,
		    akill->mask, akill->setter,
		    showreasons ? akill->reason : "");
	    else
		notice (s_OperServ, source, OS_AKILL_LIST_MATCH, j,
		    akill->mask, akill->setter,
		    duration (akill->expires - (time (NULL) - akill->set)),
		    showreasons ? akill->reason : "");
	}
    }

    notice (s_OperServ, source, OS_AKILL_LIST_END);
}

/* Check all AKills and expire any nessecary */
void expire_akills ()
{
    AutoKill *akill;
#ifndef IRCUP9
    char *user, *host;
#endif
    int i, j = 0;

    if (logupdates_on == TRUE)
	log ("Checking for expired AKills...");

    for (i = 0; i < 1024; i++)
    {
	for (akill = akilllist[i]; akill; akill = akill->next)
	{
	    /* AKills with an expire time of 0 are permanent */
	    if (akill->expires > 0)
	    {
		if ((akill->expires - (time (NULL) - akill->set)) < 0)
		{
		    log ("Expiring %sAKill on %s by %s: %s",
			akill->realname ? "Realname " : "",
			akill->mask, akill->setter, akill->reason);

#ifdef IRCUP9
		    send_cmd (me.name, "GLINE * -%s", akill->mask);
#else
		    user = sstrdup (akill->mask);
		    host = strchr (user, '@');

		    if (host)
		    {
			*host++ = 0;

			send_cmd (me.name, "%s %s %s", me.token ? "Y" :
			    "RAKILL", host, user);
		    }

		    free (user);
#endif /* IRCUP9 */

		    free (akill->setter);
		    free (akill->reason);

		    if (akill->prev)
			akill->prev->next = akill->next;
		    else
			akilllist[AHASH(akill->mask)] = akill->next;
		    if (akill->next)
			akill->next->prev = akill->prev;

		    free (akill);

		    j++;
		}
	    }
	}
    }

    /* Only log this if we have something to report, or if we should
       log it regardless.
     */
    if ((logupdates_on == FALSE && j > 0) || logupdates_on == TRUE)
	log ("Expired %d AKills.", j);
}

/* If the given information matches an AKill, kill that user and place an
   AKill on it. (If we see a user that's AKilled, then the AKill was
   probably rehashed out or the server restarted)
 */
int check_akill (const char *nick, const char *user, const char *host,
		 const char *real)
{
    AutoKill *akill;
    int i;
    char buf[512], *user2, *host2;

    strscpy (buf, user, sizeof (buf) -2);
    i = strlen (buf);
    buf[i++] = '@';
    strscpy (buf + i, host, sizeof (buf) -i);

    for (i = 0; i < 1024; i++)
	for (akill = akilllist[i]; akill; akill = akill->next)
	{
	    if ((!akill->realname && match_wild_nocase (akill->mask, buf)) ||
		 (akill->realname && match_wild_nocase (akill->mask, real)))
	    {
		user2 = sstrdup (akill->mask);
		host2 = strchr (user2, '@');

		if (!host2)
		    return 0;

		*host2++ = 0;

		send_cmd (me.name, "%s %s %s :%s", me.token ? "V" : "AKILL",
		    host2, user2, akill->reason);
		kill_user (s_OperServ, nick, akill->reason);

		free (user2);
		return 1;
	    }
	}
    return 0;
}

/* Handle a LOGONMSG command */
static void do_logonmsg (const char *source)
{
    char *cmd = strtok (NULL, " ");

    if (!cmd)
    {
	notice (s_OperServ, source, RPL_SYNTAX,
	    "LOGONMSG ADD|DEL|LIST [Text|Number]");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "LOGONMSG");
    }
    else
    {
	Hash_SUB *command, hash_table[] =
	{
	    {"ADD",		H_SRA,		do_logonmsg_add},
	    {"DEL",		H_SRA,		do_logonmsg_del},
	    {"LIST",		H_SRA,		do_logonmsg_list},
	    {NULL}
	};

	if ((command = get_sub_hash (s_OperServ, source, strupper (cmd),
				 hash_table)))
	    (*command->process) (source);
	else 
	{
	    notice (s_OperServ, source, RPL_SYNTAX,
		"LOGONMSG ADD|DEL|LIST [Text|Number]");
	    notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "LOGONMSG");
	}
    }   
}

/* Add a logonmsg to the list */
static void do_logonmsg_add (const char *source)
{
    char *msg = strtok (NULL, "");
    char buf[512];

    if (!msg)
    {
	notice (s_OperServ, source, RPL_SYNTAX, "LOGONMSG ADD Text");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "LOGONMSG");
	return;
    }

    if (strlen (msg) > 256)
    {
	notice (s_OperServ, source, ERR_MSG_LEN, strlen (msg));
	return;
    }

    if (stamplogons_on == TRUE)
    {
	time_t t;
	struct tm tm;

	time (&t);
	tm = *localtime (&t);
	strftime (buf, sizeof (buf) - 1, "[\2News\2 - %B %d, %Y] ", &tm);
	strcat (buf, msg);
    }
    else
	snprintf (buf, sizeof (buf), msg);

    noticeall (buf);
    new_message (source, time (NULL), buf);
    notice (s_OperServ, source, OS_LOGONMSG_ADDED);
}

/* Remove a logonmsg. */
static void do_logonmsg_del (const char *source)
{
    MSG *msg;
    char *param = strtok (NULL, " ");
    int i = 0;

    if (!param)
    {
	notice (s_OperServ, source, RPL_SYNTAX, "LOGONMSG DEL Number");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "LOGONMSG");
	return;
    }

    if (!stricmp (param, "ALL"))
    {
	for (msg = msglist; msg; msg = msg->next)
	{
	    i++;
	    free (msg->text);
	    free (msg->setter);

	    if (msg->prev)
		msg->prev->next = msg->next;
	    else
		msglist = msg->next;
	    if (msg->next)
		msg->next->prev = msg->prev;
	}

	free (msg);
	notice (s_OperServ, source, OS_ALL_LOGONMSG_DELETED, i);
	return;
    }
    else
    {
	for (msg = msglist; msg; msg = msg->next)
	{
	    i++;
	    if (i == atoi (param))
	    {
		free (msg->text);
		free (msg->setter);

		if (msg->prev)
		    msg->prev->next = msg->next;
		else
		    msglist = msg->next;
		if (msg->next)
		    msg->next->prev = msg->prev;

		free (msg);
		notice (s_OperServ, source, OS_LOGONMSG_DELETED, i);
		return;
	    }
	}
    }
    notice (s_OperServ, source, OS_LOGONMSG_NOT_FOUND, param);
}

/* List the current logonmsgs */
static void do_logonmsg_list (const char *source)
{
    MSG *msg;
    int i = 0;

    notice (s_OperServ, source, OS_LOGONMSG_LIST_START);

    for (msg = msglist; msg; msg = msg->next)
    {
	i++;
	notice (s_OperServ, source, OS_LOGONMSG_LIST, i, msg->setter,
	    time_ago (msg->set));
	notice (s_OperServ, source, msg->text);
    }

    notice (s_OperServ, source, OS_LOGONMSG_LIST_END);
}

/* Display the list of clones on the network */
static void do_clones (const char *source)
{
    Clone *clone;
    User *user;
    char *param = strtok (NULL, " ");
    char *limit = strtok (NULL, " ");
    int i, j, max = 250, shown = 0;

    if (!param)
    {
	notice (s_OperServ, source, RPL_SYNTAX, "CLONES Pattern [Limit]");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "CLONES");
	return;
    }

    if (limit)
    {
	max = atoi (limit);

	if (max > 250)
	    max = 250;
    }

    notice (s_OperServ, source, OS_CLONES_START, max);
    notice (s_OperServ, source, OS_CLONES_HEADER);

    for (i = 0; i < 1024; i++)
    {
	for (clone = clonelist[i]; clone; clone = clone->next)
	{
	    int shost = 0;

	    if (!match_wild_nocase (param, clone->host))
		continue;

	    if (clone->cnt < 2)
		continue;

	    for (j = 0; j < 1024; j++)
	    {
		for (user = userlist[j]; user; user = user->next)
		{
		    /* Very few hosts will go over this.. */
		    notice (s_OperServ, source, "%s%-44s %s%s%s%s%s",
			shost ? "" : "\2", shost ? "" : clone->host,
			user->nick, shost ? "" : " (",
			shost ? "" : itoa (clone->cnt), shost ? "" : ")",
			shost ? "" : "\2");
		    shost = 1;
		    shown++;
		}
	    }
	}
    }

    notice (s_OperServ, source, OS_CLONES_END, shown);
}

/* Handle a TRIGGER command */
static void do_trigger (const char *source)
{
    char *cmd = strtok (NULL, " ");

    if (!cmd)
    {
	notice (s_OperServ, source, RPL_SYNTAX,
	    "TRIGGER ADD|DEL|LIST [Host|Number] [Limit]");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "TRIGGER");
    }

    else
    {
	Hash_SUB *command, hash_table[] =
	{
	    {"ADD",		H_IRCOP,	do_trigger_add},
	    {"DEL",		H_IRCOP,	do_trigger_del},
	    {"LIST",		H_IRCOP,	do_trigger_list},
	    {NULL}
	};

	if ((command = get_sub_hash (s_OperServ, source, strupper (cmd),
				 hash_table)))
	    (*command->process) (source);
	else
	{
	    notice (s_OperServ, source, RPL_SYNTAX,
		"TRIGGER ADD|DEL|LIST [Host|Number] [Limit]");
	    notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "TRIGGER");
	}
    }
}

/* Add a trigger to the list */
static void do_trigger_add (const char *source)
{
    Trigger *trigger;
    char *mask = strtok (NULL, " "), *limit = strtok (NULL, " ");
    char *username, *hostname;

    if (!mask || !limit)
    {
	notice (s_OperServ, source, RPL_SYNTAX, "TRIGGER ADD Host Limit");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "TRIGGER");
	return;
    }

    if (!strchr (mask, '@'))
    {
	notice (s_OperServ, source, OS_TRIGGER_FORMAT);
	return;
    }

    if (!is_sra (source))
    {
	int i, nonwild = 0;

	for (i = 0; i < strlen (mask); i++)
	    if (!(mask[i] == '*' || mask[i] == '?' || mask[i] == '@' ||
		mask[i] == '.'))
		nonwild++;

	if (nonwild < nonwildreq)
	{
	    notice (s_OperServ, source, ERR_TOOMANYWILD, nonwildreq);
	    return;
	}
    }

    username = sstrdup (mask);
    hostname = strchr (username, '@');

    if (!hostname)
	return;

    *hostname++ = 0;

    if ((trigger = findtrig (username, hostname)))
    {
	trigger->limit = atoi (limit);
	notice (s_OperServ, source, OS_RETRIGGER, mask, atoi (limit));
	free (username);
	return;
    }

    free (username);

    new_trigger (mask, atoi (limit));
    notice (s_OperServ, source, OS_TRIGGER_ADDED, mask);
}

/* Remove a trigger. */
static void do_trigger_del (const char *source)
{
    Trigger *trigger;
    char *param = strtok (NULL, " ");
    int i = 0;

    if (!param)
    {
	notice (s_OperServ, source, RPL_SYNTAX, "TRIGGER DEL Number|Host|ALL");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "TRIGGER");
	return;
    }

    if (!stricmp (param, "ALL"))
    {
	for (trigger = triggerlist; trigger; trigger = trigger->next)
	{
	    i++;

	    /* Notice them up here so we can show the host */
	    notice (s_OperServ, source, OS_ALL_TRIGGER_DELETED, trigger->host);

	    if (trigger->prev)
		trigger->prev->next = trigger->next;
	    else
		triggerlist = trigger->next;
	    if (trigger->next)
		trigger->next->prev = trigger->prev;
	}

	free (trigger);
	return;
    }
    else
    {
	for (trigger = triggerlist; trigger; trigger = trigger->next)
	{
	    i++;
	    if (!stricmp (param, trigger->host) || i == atoi (param))
	    {
		if (trigger->prev)
		    trigger->prev->next = trigger->next;
		else
		    triggerlist = trigger->next;
		if (trigger->next)
		    trigger->next->prev = trigger->prev;

		free (trigger);
		if (atoi (param) > 0)
		    notice (s_OperServ, source, OS_TRIGGER_DELETED, "#", param);
		else
		    notice (s_OperServ, source, OS_TRIGGER_DELETED, "on", param);
		return;
	    }
	}
    }
    notice (s_OperServ, source, OS_TRIGGER_NOT_FOUND, param);
}

/* List the current triggers */
static void do_trigger_list (const char *source)
{
    Trigger *trigger;
    int i = 0;

    notice (s_OperServ, source, OS_TRIGGER_LIST_START);

    for (trigger = triggerlist; trigger; trigger = trigger->next)
    {
	i++;
	notice (s_OperServ, source, OS_TRIGGER_LIST, i, trigger->host,
	    trigger->limit);
    }

    notice (s_OperServ, source, OS_TRIGGER_LIST_END);
}

/* Handle a EXCEPTION command */
static void do_exception (const char *source)
{
    char *cmd = strtok (NULL, " ");

    if (!cmd)
    {
	notice (s_OperServ, source, RPL_SYNTAX,
	    "EXCEPTION ADD|DEL|LIST [Host|Number]");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "EXCEPTION");
    }

    else
    {
	Hash_SUB *command, hash_table[] =
	{
	    {"ADD",		H_IRCOP,	do_exception_add},
	    {"DEL",		H_IRCOP,	do_exception_del},
	    {"LIST",		H_IRCOP,	do_exception_list},
	    {NULL}
	};

	if ((command = get_sub_hash (s_OperServ, source, strupper (cmd),
				 hash_table)))
	    (*command->process) (source);
	else
	{
	    notice (s_OperServ, source, RPL_SYNTAX,
		"EXCEPTION ADD|DEL|LIST [Host|Number]");
	    notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "EXCEPTION");
	}
    }
}

/* Add a exception to the list */
static void do_exception_add (const char *source)
{
    Exception *exception;
    char *mask = strtok (NULL, " ");
    char *username, *hostname;

    if (!mask)
    {
	notice (s_OperServ, source, RPL_SYNTAX, "EXCEPTION ADD Host");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "EXCEPTION");
	return;
    }

    if (!strchr (mask, '@'))
    {
	notice (s_OperServ, source, OS_EXCEPTION_FORMAT);
	return;
    }

    if (!is_sra (source))
    {
	int i, nonwild = 0;

	for (i = 0; i < strlen (mask); i++)
	    if (!(mask[i] == '*' || mask[i] == '?' || mask[i] == '@' ||
		mask[i] == '.'))
		nonwild++;

	if (nonwild < nonwildreq)
	{
	    notice (s_OperServ, source, ERR_TOOMANYWILD, nonwildreq);
	    return;
	}
    }

    username = sstrdup (mask);
    hostname = strchr (username, '@');

    if (!hostname)
	return;

    *hostname++ = 0;

    if ((exception = findexcept (username, hostname)))
    {
	notice (s_OperServ, source, OS_ALREADY_EXCEPT, mask);
	free (username);
	return;
    }

    free (username);

    new_exception (mask);
    notice (s_OperServ, source, OS_EXCEPTION_ADDED, mask);
}

/* Remove a exception. */
static void do_exception_del (const char *source)
{
    Exception *exception;
    char *param = strtok (NULL, " ");
    int i = 0;

    if (!param)
    {
	notice (s_OperServ, source, RPL_SYNTAX, "EXCEPTION DEL Number|Host|ALL");
	notice (s_OperServ, source, ERR_MORE_INFO, s_OperServ, "EXCEPTION");
	return;
    }

    if (!stricmp (param, "ALL"))
    {
	for (exception = exceptionlist; exception; exception = exception->next)
	{
	    i++;

	    /* Notice them up here so we can show the host */
	    notice (s_OperServ, source, OS_EXCEPTIONS_DELETED,
		exception->host);

	    if (exception->prev)
		exception->prev->next = exception->next;
	    else
		exceptionlist = exception->next;
	    if (exception->next)
		exception->next->prev = exception->prev;
	}

	free (exception);
	return;
    }
    else
    {
	for (exception = exceptionlist; exception; exception = exception->next)
	{
	    i++;
	    if (!stricmp (param, exception->host) || i == atoi (param))
	    {
		if (exception->prev)
		    exception->prev->next = exception->next;
		else
		    exceptionlist = exception->next;
		if (exception->next)
		    exception->next->prev = exception->prev;

		free (exception);
		if (atoi (param) > 0)
		    notice (s_OperServ, source, OS_EXCEPTION_DELETED, "#", param);
		else
		    notice (s_OperServ, source, OS_EXCEPTION_DELETED, "on", param);
		return;
	    }
	}
    }
    notice (s_OperServ, source, OS_EXCEPTION_NOT_FOUND, param);
}

/* List the current exceptions */
static void do_exception_list (const char *source)
{
    Exception *exception;
    int i = 0;

    notice (s_OperServ, source, OS_EXCEPTION_LIST_START);

    for (exception = exceptionlist; exception; exception = exception->next)
    {
	i++;
	notice (s_OperServ, source, OS_EXCEPTION_LIST, i, exception->host);
    }

    notice (s_OperServ, source, OS_EXCEPTION_LIST_END);
}

/* Load the OperServ DB from disk */
void load_os_dbase (void)
{
    FILE *f = fopen (OPERSERV_DB, "r");
    char *item, dBuf[2048], backup [2048];

    log ("Loading %s...", OPERSERV_DB);

    if (!f)
    {
	log ("No %s found!", OPERSERV_DB);

	/* Check for a db.save */
	strcpy (backup, OPERSERV_DB);
	strcat (backup, ".save");
	if (rename (backup, OPERSERV_DB) > 0)
	{
	    log ("Recovering %s.save", OPERSERV_DB);
	    f = fopen (OPERSERV_DB, "r");
	}
	else
	    return;
    }

    while (fgets (dBuf, 2047, f))
    {
	item = strtok (dBuf, " ");

	/* Ignore certain things.. */
	if (*item == '#' || *item == '\n' || *item == '\t' || *item == ' ' ||
	    *item == '\0' || *item == '\r')
	    continue;

	if (!strcmp (item, "AK"))
	{
	    AutoKill *akill, **list;

	    akill = scalloc (sizeof (AutoKill), 1);
	    akill->mask = sstrdup (strtok (NULL, " "));
	    akill->setter = sstrdup (strtok (NULL, " "));
	    akill->realname = atoi (strtok (NULL, " "));
	    akill->set = atol (strtok (NULL, " "));
	    akill->expires = atol (strtok (NULL, " "));
	    akill->reason = sstrdup (strtok (NULL, "\n"));
	    list = &akilllist[AHASH(akill->mask)];
	    akill->next = *list;
	    if (*list)
		(*list)->prev = akill;
	    *list = akill;
	}

	if (!strcmp (item, "LM"))
	{
	    char *setter, *text;
	    time_t set;

	    setter = strtok (NULL, " ");
	    set = atol (strtok (NULL, " "));
	    text = strtok (NULL, "\n");

	    new_message (setter, set, text);
	}

	if (!strcmp (item, "TR"))
	{
	    char *host;
	    int limit;

	    host = strtok (NULL, " ");
	    limit = atoi (strtok (NULL, "\n"));

	    new_trigger (host, limit);
	}

	if (!strcmp (item, "EX"))
	{
	    char *host;

	    host = strtok (NULL, "\n");

	    new_exception (host);
	}
    }
    fclose (f);
}

/* Write the operserv db to disk. */
void save_os_dbase ()
{
    FILE *f;
    AutoKill *akill;
    MSG *msg;
    Trigger *trigger;
    Exception *exception;
    int i;
    char backup[2048];

    if (logupdates_on == TRUE)
	log ("Saving %s...", OPERSERV_DB);

    strcpy (backup, OPERSERV_DB);
    strcat (backup, ".save");
    remove (backup);
    if (rename (OPERSERV_DB, backup) < 0)
	log ("Can't back up %s!", OPERSERV_DB);
    f = fopen (OPERSERV_DB, "w");

    if (!f)
    {
	log ("Can't write to %s!", OPERSERV_DB);
	if (rename (backup, OPERSERV_DB) < 0)
	    fatal ("Can't restore backup of %s!", OPERSERV_DB);
	return;
    }

    /* Print an identifying line at the top of the DB. */
    fprintf (f, "OperStats v%s %s\n", version, OPERSERV_DB);

    /* Write any AKills: One line per AKill. Format:
       AK mask setter realname? set expires reason
     */
    for (i = 0; i < 1024; i++)
	for (akill = akilllist[i]; akill; akill = akill->next)
	    fprintf (f, "AK %s %s %d %lu %lu %s\n", akill->mask, akill->setter,
		akill->realname, akill->set, akill->expires, akill->reason);

    /* Write any LogonMSGs: One line per message. Format:
       LM setter set text
     */
    for (msg = msglist; msg; msg = msg->next)
	fprintf (f, "LM %s %lu %s\n", msg->setter, msg->set, msg->text);

    /* Write any Triggers: One line per trigger. Format:
       TR host limit
     */
    for (trigger = triggerlist; trigger; trigger = trigger->next)
	fprintf (f, "TR %s %d\n", trigger->host, trigger->limit);

    /* Write any Exceptions: One line per exception. Format:
       EX host
     */
    for (exception = exceptionlist; exception; exception = exception->next)
	fprintf (f, "EX %s\n", exception->host);

    fclose (f);
    remove (backup);
}

/* Shared functions: These call shared functions with the appropriate
   information.
 */
static void sh_auth (const char *source)
{
    do_auth (s_OperServ, source);
}

static void sh_deauth (const char *source)
{
    do_deauth (s_OperServ, source);
}

#ifdef RAWINJECT
static void sh_raw (const char *source)
{
    do_raw (s_OperServ, source);
}

static void sh_inject (const char *source)
{
    do_inject (s_OperServ, source);
}
#endif

static void sh_debug (const char *source)
{
    do_debug (s_OperServ, source);
}

static void sh_shutdown (const char *source)
{
    do_shutdown (s_OperServ, source);
}

static void sh_restart (const char *source)
{
    do_restart (s_OperServ, source);
}

static void sh_settings (const char *source)
{
    do_settings (s_OperServ, source);
}

static void sh_update (const char *source)
{
    do_update (s_OperServ, source);
}

static void sh_rehash (const char *source)
{
    do_rehash (s_OperServ, source);
}

/* Every time I think I've got the channels/user counts to stop
   getting desync'd from the network, I take these out .. and every
   time I do, some other new method of getting desync'd pops up. Feh.
 */
static void do_userlist (const char *source)
{
    User *u;
    int i, j = 0;
            
    for (i = 0; i < 1024; i++)
    {
        for (u = userlist[i]; u; u = u->next)
        {
            j++;
            notice (s_OperServ, source, "%s (%s@%s)", u->nick,
                u->user, u->host);
        }
    }
    notice (s_OperServ, source, "%d users", j);
}
            
static void do_chanlist (const char *source)
{
    Channel *c;
    int i, j = 0;
        
    for (i = 0; i < 1024; i++)
    {
        for (c = chanlist[i]; c; c = c->next)
        {
            j++;
            notice (s_OperServ, source, "%s %s%s%s", c->name,
                c->mode & CMODE_s ? "(" : "",
                c->mode & CMODE_s ? "s" : "",
                c->mode & CMODE_s ? ")" : "");
        }
    }
    notice (s_OperServ, source, "%d channels", j);
}
